use std::collections::HashMap;
use std::env;
use std::fs::read_to_string;

use clap::{Parser, Subcommand};
use ed25519_dalek::{Signature, SigningKey, VerifyingKey};
use hex;
use sqlx::sqlite::SqlitePool;

#[derive(Parser)]
struct Args {
    #[command(subcommand)]
    cmd: Option<Command>,
}

#[derive(Subcommand)]
enum Command {
    Add {
        ephemeral_key: String,
        persistent_key: String,
        expires_at: i64,
        signature: String,
    },
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    let args = Args::parse();
    let keys_path = env::var("CLL2V0_KEYS").expect("Expect `CLL2V0_KEYS` to be set");
    // let db_url = env::var("CLL2V0_DB").expect("Expect `CLL2V0_DB` to be set");
    let db_url = env::var("DATABASE_URL").expect("Expect `DATABASE_URL` to be set");
    let signing_keys = read_to_string(&keys_path).expect("Error reading file");

    let pool = SqlitePool::connect(&db_url).await?;

    let mut keychain: HashMap<[u8; 32], SigningKey> = HashMap::new();
    for key in signing_keys.lines() {
        let secret_key = hex::decode(key.trim()).unwrap().try_into().unwrap();
        let skey = SigningKey::from_bytes(&secret_key);
        let vkey = skey.verifying_key();
        keychain.insert(vkey.to_bytes(), skey);
        let _ = add_persistent_key(&pool, vkey).await;
    }

    match args.cmd {
        Some(Command::Add {
            ephemeral_key,
            persistent_key,
            expires_at,
            signature,
        }) => {
            println!("Adding new ephemeral key '{ephemeral_key}'");
            let e = VerifyingKey::from_bytes(
                &hex::decode(ephemeral_key)
                    .expect("Cannot parse ephemeral key")
                    .try_into()
                    .expect("wrong length"),
            )
            .expect("bad key");
            let p = VerifyingKey::from_bytes(
                &hex::decode(persistent_key)
                    .expect("Cannot parse persistent")
                    .try_into()
                    .expect("wrong length"),
            )
            .expect("bad key");
            let s = Signature::from_bytes(
                &hex::decode(signature)
                    .expect("Cannot parse persistent")
                    .try_into()
                    .expect("wrong length"),
            );

            let res = add_ephemeral_key(&pool, &e, &p, expires_at, &s).await?;
            println!("Added new ekey: {}", res);
        }
        None => {
            println!("Printing list of all todos");
            list_persistent_keys(&pool).await?;
        }
    }

    Ok(())
}

async fn add_persistent_key(pool: &SqlitePool, vkey: VerifyingKey) -> anyhow::Result<i64> {
    let mut conn = pool.acquire().await?;
    let b = Into::<Vec<u8>>::into(vkey.to_bytes());

    // Insert the task, then obtain the ID of this row
    let id = sqlx::query!(
        r#"
INSERT INTO persistent_keys ( id )
VALUES ( ?1 )
        "#,
        b
    )
    .execute(&mut *conn)
    .await?
    .last_insert_rowid();

    Ok(id)
}

async fn add_ephemeral_key(
    pool: &SqlitePool,
    ephemeral_key: &VerifyingKey,
    persistent_key: &VerifyingKey,
    expires_at: i64,
    signature: &Signature,
) -> anyhow::Result<i64> {
    let mut conn = pool.acquire().await?;

    let ephemeral_key_ = Into::<Vec<u8>>::into(ephemeral_key.to_bytes());
    let persistent_key_ = Into::<Vec<u8>>::into(persistent_key.to_bytes());
    let signature_ = Into::<Vec<u8>>::into(signature.to_bytes());
    // Insert the task, then obtain the ID of this row
    let res = sqlx::query!(
        r#"
INSERT INTO ephemeral_keys (
    id,
    persistent_key,
    expires_at,
    signature
)
VALUES ( ?1, ?2, ?3, ?4 )
        "#,
        ephemeral_key_,
        persistent_key_,
        expires_at,
        signature_,
    )
    .execute(&mut *conn)
    .await?
    .last_insert_rowid();

    Ok(res)
}

async fn list_persistent_keys(pool: &SqlitePool) -> anyhow::Result<()> {
    let recs = sqlx::query!(
        r#"
SELECT id
FROM persistent_keys
ORDER BY id
        "#
    )
    .fetch_all(pool)
    .await?;

    for rec in recs {
        println!("- [{}]", hex::encode(rec.id),);
    }

    Ok(())
}

fn print_key() {
    // Useful for setting up some keys
    let mut rng = rand::rngs::OsRng;
    println!("{}", hex::encode(SigningKey::generate(&mut rng).to_bytes()));
}
